# encoding: utf-8
from __future__ import unicode_literals
import abc
from datetime import datetime

from django.db.models import Q

from edu_to_analytics.service.utils import PERIOD_TODAY, PERIOD_PERIOD


# Отношение модель - название поля через которое можно достать
# информацию о прикреплении к чреждению
UNIT_FIELD_MAPPING = {
    'Pupil': 'grup__unit_id',
    'Group': 'unit_id',
    'Children': 'declaration__declarationunit__unit_id',
    'Delegate': (
        'childrendelegate__'
        'children__'
        'declaration__'
        'declarationunit__'
        'unit_id'),
    'Direct': 'group__unit_id',
    'Declaration': 'declarationunit__unit_id'
}


class CalculatedIndex(object):
    """Класс для подсчитываемых показателей из других показателей."""
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def calculate(self, unit, period):
        """Единственный обязательный метод который нужно перебить."""


class ModelIndex(object):
    """Класс показателя который работает с моделью."""
    __metaclass__ = abc.ABCMeta

    #: Один или несколько классов для обработки логов
    logs = []
    #: Дополнительная фильтрация.
    #: Фильтр по учреждению идет изначально
    extra_filter = Q()
    #: Дополнительный параметр для подсчетов, пока используется
    #: только в подсчете двух показателей детей в очереди
    extra_param = None
    #: Поле для получения айдишников объектов по которым будет идти
    #: поиск изменений в логах
    field = 'id'

    @abc.abstractproperty
    def model(self):
        """Модель с которой работаем."""

    @property
    def unit_relation(self):
        """Возвращает поля по которому можно найти связь с учреждением."""
        return UNIT_FIELD_MAPPING.get(self.model.__name__)

    def qs(self, unit):
        """
        Возвращает все записи принадлежащие учреждению.

        Используется для переопределнных вызовов calculate,
        когда необходимо например, отдать сумму значений по полю
        вместо подсчета записей в базе
        """
        if not self.unit_relation:
            raise KeyError('Необходимо обновить маппинг')
        return self.model.objects.filter(**{self.unit_relation: unit['id']})

    def calculate(self, unit, period):
        """Основной метод который отдает результат вычислений."""
        # Подсчет на текущий день, рассчитывается при любом периоде
        # потому что потом это значение используется для вычисления
        # показателя на вчера и период дат
        if self.model.__name__ == 'Delegate':
            object_ids = set(
                list(self.model.objects.filter(
                    childrendelegate__children__declaration__declarationunit__unit_id=unit['id']
                ).filter(self.extra_filter).values_list('id', flat=True)) +
                list(self.model.objects.filter(
                    childrendelegate__children__pupil__grup__unit_id=unit['id']
                ).filter(self.extra_filter).values_list(self.field, flat=True)))
        else:
            object_ids = set(
                self.model.objects
                .filter(self.extra_filter, **{self.unit_relation: unit['id']})
                .values_list(self.field, flat=True))
        total = len(object_ids)

        report_is_for_today = (
            period.type == PERIOD_TODAY or period.start == datetime.now())
        if not report_is_for_today:
            for log in self.logs:
                logs = log(unit, period)
                history_val = sum(logs.history_value(object_ids))
                # Аналитики решили что если подсчет по периоду
                # то мы возвращаем количество событий в этот период
                if period.type == PERIOD_PERIOD:
                    total = history_val
                # Но если это подсчет на вчера или день, то
                # вернуть надо общий показатель минус количество событий
                else:
                    total -= history_val

        return total
